package pigbarf;

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class Generator {
   private static final String RESULT = "pigbarf.AbstractParser.Result";
   private static final String POSITION = "pigbarf.AbstractParser.Position";
   private static final String IOEXCEPTION = "java.io.IOException";
   private static final String INPUTSTREAM = "java.io.InputStream";
   private static final String PARSEEXCEPTION = "pigbarf.ParsingException";
   private static final String MINIPARSER = "pigbarf.AbstractParser.MiniParser";
   private static final String TAB = "   ";
   private static final String TAB2 = TAB + TAB;
   private static final String TAB3 = TAB2 + TAB;
   private static final String TAB4 = TAB3 + TAB;
   private static final String TAB5 = TAB4 + TAB;
   private static final String TAB6 = TAB5 + TAB;
   private final Specification spec;
   private final String className;

   public Generator(Specification _spec, String _className) {
      this.spec = _spec;
      this.className = _className;
   }

   public void generateParser(PrintStream out) {
      this.printHeader(out);

      for (int i = 0; i < this.spec.getNumRules(); i++) {
         Rule rule = this.spec.getRule(i);
         this.printRuleParser(rule, out);
      }
      
      this.printStart(spec.getStartRule(), out);

      // close class
      out.println("}");
   }

   private void printHeader(PrintStream out) {
      // print headers
      if (this.spec.getHeaders() != null)
         out.println(this.spec.getHeaders());
      
      // print class line
      out.println("public class " + this.className + " extends pigbarf.AbstractParser {");
      out.println(TAB + "public " + this.className + "(" + INPUTSTREAM + " _input) {");
      out.println(TAB2 + "super(_input, " + this.spec.isByteOrderBE() + ", " + this.spec.isBitOrderBE() + ");");
      out.println(TAB + "}");
      out.println(TAB + "public " + this.className + "(byte[] _input) {");
      out.println(TAB2 + "super(_input, " + this.spec.isByteOrderBE() + ", " + this.spec.isBitOrderBE() + ");");
      out.println(TAB + "}");
   }

   private String escapeString(String str) {
      StringBuilder result = new StringBuilder(str.length());
      for (int i = 0; i < str.length(); i++) {
         char c = str.charAt(i);
         switch (c) {
         case '\\':
            result.append("\\\\");
            break;
         case '\n':
            result.append("\\n");
            break;
         case '\t':
            result.append("\\t");
            break;
         case '\r':
            result.append("\\r");
            break;
         case '"':
            result.append("\\\"");
            break;

         default:
            result.append(c);
            break;
         }
      }
      return result.toString();
   }

   private void printRuleParser(Rule rule, PrintStream out) {
      final Type ruleType = rule.getType();

      final String returnType;
      final String wrappedType;
      if (rule.getNumParameters() == 0) {
         returnType = ruleType.getSimpleSelf().getType();
         wrappedType = ListRuleItem.wrapType(returnType);
         // simple rule
         out.println(TAB + "protected final " + MINIPARSER + "<" + wrappedType + "> _yy_rule_" + rule.getName() + " = ");
         out.println(TAB2 + "new " + MINIPARSER + "<" + wrappedType + ">() {");
      } else {
         FunctionType funcType = ruleType.getFunctionSelf();
         returnType = funcType.getReturnType().getType();
         wrappedType = ListRuleItem.wrapType(returnType);

         // parameterized rule
         out.println(TAB + "protected class _yy_rule_" + rule.getName() + " implements " + MINIPARSER + "<" + wrappedType + "> {");
         for (int i = 0; i < funcType.getNumParameters(); i++) {
            out.println(TAB2 + "private final " + funcType.getParameterType(i) + " " + rule.getFormalName(i) + ";");
         }

         // constructor header
         out.print(TAB2 + "public _yy_rule_" + rule.getName() + "(");
         for (int i = 0; i < funcType.getNumParameters(); i++) {
            if (i>0) out.print(", ");
            out.print(funcType.getParameterType(i) + " _" + rule.getFormalName(i));
         }
         out.println(") {");

         // assign fields
         for (int i = 0; i < funcType.getNumParameters(); i++) {
            out.println(TAB3 + "this." + rule.getFormalName(i) + " = _" + rule.getFormalName(i) + ";");
         }
         
         // end constructor
         out.println(TAB2 + "}");
      }

      ////////////// SHARED CODE ////////////
      
      // print options
      for (int i = 0; i < rule.getNumOptions(); i++) {
         this.printOption(rule, i, out);
      }
      
      // begin get method
      out.println(TAB3 + "public Result<" + wrappedType + "> get() throws " + IOEXCEPTION + ", " + PARSEEXCEPTION + " {");
      
      if (rule.getNumOptions() == 1) {
         out.println(TAB4 + POSITION + " _yy_startPosition = _yy_getPosition();");
         out.println(TAB4 + RESULT + "<" + wrappedType + "> result = option1();");
         out.println(TAB4 + "if (!result.failed()) return result;");
         out.println(TAB4 + "else {");
         out.println(TAB5 + "_yy_setPosition(_yy_startPosition);");
         out.println(TAB5 + "return _yy_fail(\"Error parsing rule " + rule.getName() + "\", result);");
         out.println(TAB4 + "}");
      } else {
         out.println(TAB4 + POSITION + " _yy_startPosition = _yy_getPosition();");
         out.println(TAB4 + RESULT + "<" + wrappedType + "> result = option1();");
         out.println(TAB4 + "if (!result.failed()) return result;");
         for (int i = 1; i < rule.getNumOptions(); i++) {
            out.println(TAB4 + "_yy_setPosition(_yy_startPosition);");
            out.println(TAB4 + "result = option" + (i+1) + "();");
            out.println(TAB4 + "if (!result.failed()) return result;");
         }
         out.println(TAB4 + "_yy_setPosition(_yy_startPosition);");
         out.println(TAB4 + "return _yy_fail(\"All options failed for rule " + rule.getName() + "\");");
      }
      
      // end get method
      out.println(TAB3 + "}");

      //////////// END OF SHARED CODE /////////
      
      if (rule.getNumParameters() == 0) {
         // end class defn
         out.println(TAB2 + "};");
      } else {
         // end class defn
         out.println(TAB + "}");
      }
   }

   private void printOption(Rule rule, int option, PrintStream out) {
      Type type = rule.getType();
      RHS rhs = rule.getOption(option);

      String returnType;
      if (type.isSimple()) {
         // simple type
         returnType = type.getSimpleSelf().getType();
      } else {
         // function type
         returnType = type.getFunctionSelf().getReturnType().getType();
      }
      String wrappedType = ListRuleItem.wrapType(returnType);
      
      out.println(TAB3 + "private " + RESULT + "<" + wrappedType + "> option" + (option+1) + "() throws " + IOEXCEPTION + ", " + PARSEEXCEPTION + " {");
      out.println(TAB4 + POSITION + " _yy_startPosition = _yy_getPosition();");
      if (returnType.equals("void")) {
         out.println(TAB4 + wrappedType + " _yy_0 = null;");
      } else {
         out.println(TAB4 + returnType + " _yy_0;");
      }
      int valuecount = 0;
      for (int i = 0; i < rhs.getNumRuleItems(); i++) {
         RuleItem item = rhs.getRuleItem(i);
         if (item.isValue()) {
            this.printValueItem(rule.getName(), item.getValueSelf(), valuecount, out);
            valuecount++;
         } else if (item.isPredicate()) {
            PredicateRuleItem predicate = item.getPredicateSelf();
            String code = replaceDollars(predicate.getPredicateString());
            out.println(TAB4 + "if (!(" + code + "))");
            out.println(TAB5 + "return _yy_fail(\"Predicate failed while parsing rule " + rule.getName() + ": " + escapeString(code) + "\");");
         } else if (item.isCode()) {
            CodeRuleItem code = item.getCodeSelf();
            String str = replaceDollars(code.getCodeString());
            out.println(TAB4 + str);
         } else if (item.isAlign()) {
            AlignRuleItem align = item.getAlignSelf();
            String str = replaceDollars(align.getBoundaryString());
            out.println(TAB4 + "if (!_yy_align(" + str + "))");
            out.println(TAB5 + "return _yy_fail(\"Cannot align to byte boundary in rule " + rule.getName() + "\");");
         } else if (item.isEOF()) {
            out.println(TAB4 + "if (!_yy_atEOF())");
            out.println(TAB5 + "return _yy_fail(\"Extra bytes found after EOF rule\");");
         } else {
            throw new RuntimeException("Unrecognized rule item: " + item);
         }
      }
      out.println(TAB4 + "return _yy_success(_yy_0);");
      out.println(TAB3 + "}");
   }

   private void printValueItem(String ruleName, ValueRuleItem item, int valuecount, PrintStream out) {
      String returnType = item.getType().getType();
      String wrappedType = ListRuleItem.wrapType(returnType);
      
      if (returnType.equals("void"))
         out.println(TAB4 + wrappedType + " _yy_" + (valuecount+1) + ";");
      else
         out.println(TAB4 + returnType + " _yy_" + (valuecount+1) + ";");
      
      if (item.isSwitch()) {
         // do if-else chain for switch
         SwitchRuleItem sw = item.getSwitchSelf();
         printSwitch(sw, ruleName, valuecount+1, out);
      } else if (item.isExpression()) {
         // just assign expression to variable
         ExpressionRuleItem expression = item.getExpressionSelf();
         out.println(TAB4 + "{");
         out.println(TAB5 + "_yy_" + (valuecount+1) + " = " + replaceDollars(expression.getExpressionString()) + " ;");
         out.println(TAB4 + "}");
      } else if (item.isSubparse()) {
         out.println(TAB4 + "{");

         out.println(TAB5 + "_yy_startPosition = _yy_getPosition();");
         // make parser
         out.println(TAB5 + MINIPARSER + "<" + wrappedType + "> parser = ");
         out.println(TAB6 + parserConstructor(item) + ";");
         // make result
         out.println(TAB5 + RESULT + "<" + wrappedType + "> result = parser.get();");
         out.println(TAB5 + "if (result.failed()) {");
         out.println(TAB6 + "_yy_setPosition(_yy_startPosition);");
         out.println(TAB6 + "return _yy_fail(\"Subparse failed in rule " + ruleName + "\", result);");
         out.println(TAB5 + "}");
         out.println(TAB5 + "_yy_" + (valuecount+1) + " = result.result();");

         out.println(TAB4 + "}");
      } else {
         out.println(TAB4 + "{");

         // make parser
         out.println(TAB5 + MINIPARSER + "<" + wrappedType + "> parser = ");
         out.println(TAB6 + parserConstructor(item) + ";");
         // make result
         out.println(TAB5 + RESULT + "<" + wrappedType + "> result = parser.get();");
         out.println(TAB5 + "if (result.failed())");
         out.println(TAB6 + "return _yy_fail(result);");
         out.println(TAB5 + "_yy_" + (valuecount+1) + " = result.result();");

         out.println(TAB4 + "}");
      }
   }

   private void printSwitch(SwitchRuleItem sw, String ruleName, int valueindex, PrintStream out) {
      final String returnType = sw.getType().getType();
      final String wrappedType = ListRuleItem.wrapType(returnType);
      final String target = replaceDollars(sw.getTargetString());
      
      for (int i = 0; i < sw.getNumCases(); i++) {
         SwitchRuleItem.Case c = sw.getCase(i);
         String match = replaceDollars(c.match);
         if (i==0) {
            out.println(TAB4 + "if (_yy_checkvalue(" + target + ", " + match + ")) {");
         } else {
            out.println(TAB4 + "else if (_yy_checkvalue(" + target + ", " + match + ")) {");
         }

         if (c.rule.isExpression()) {
            // expressions have no parser
            out.println(TAB5 + "_yy_" + valueindex + " = " + replaceDollars(c.rule.getExpressionSelf().getExpressionString()) + " ;");
         } else {
            // make parser
            out.println(TAB5 + MINIPARSER + "<" + wrappedType + "> parser = ");
            out.println(TAB6 + parserConstructor(c.rule) + ";");
            
            // call parser
            out.println(TAB5 + RESULT + "<" + wrappedType + "> result = parser.get();");
            out.println(TAB5 + "if (result.failed())");
            out.println(TAB6 + "return _yy_fail(\"Switch option failed in rule " + ruleName + ": [[" + escapeString(c.match) + "]]\", result);");
            out.println(TAB5 + "_yy_" + valueindex + " = result.result();");
         }
         
         out.println(TAB4 + "}");
      }

      if (sw.getDefaultCase() != null) {
         out.println(TAB4 + "else {");

         if (sw.getDefaultCase().isExpression()) {
            // expressions have no parser
            out.println(TAB5 + "_yy_" + valueindex + " = " + replaceDollars(sw.getDefaultCase().getExpressionSelf().getExpressionString()) + " ;");
         } else {
            // make and call parser
            out.println(TAB5 + MINIPARSER + "<" + wrappedType + "> parser = ");
            out.println(TAB6 + parserConstructor(sw.getDefaultCase()) + ";");
            
            // call parser
            out.println(TAB5 + RESULT + "<" + wrappedType + "> result = parser.get();");
            out.println(TAB5 + "if (result.failed())");
            out.println(TAB6 + "return _yy_fail(\"Switch default option failed in rule " + ruleName + "\", result);");
            out.println(TAB5 + "_yy_" + valueindex + " = result.result();");
         }

         out.println(TAB4 + "}");
      } else {
         // no matched value, failure
         out.println(TAB4 + "else {");
         out.println(TAB5 + "return _yy_fail(\"No switch case matched target\");");
         out.println(TAB4 + "}");
      }
   }

   private String parserConstructor(ValueRuleItem item) {
      if (item.isNonterminal()) {
         NonterminalRuleItem nonterminal = item.getNonterminalSelf();
         if (nonterminal.getNumActuals() == 0) {
            // simple
            return "_yy_rule_" + nonterminal.getName();
         } else {
            // function
            StringBuilder builder = new StringBuilder("new _yy_rule_" + nonterminal.getName() + "(");
            for (int i = 0; i < nonterminal.getNumActuals(); i++) {
               if (i>0) builder.append(", ");
               builder.append(replaceDollars(nonterminal.getActual(i)));
            }
            return builder.append(")").toString();
         }
      } else if (item.isPrimitive()) {
         PrimitiveRuleItem primitive = item.getPrimitiveSelf();
         String name = "_yy_" + primitive.getPrimitive().name().toLowerCase();
         StringBuilder constructor = new StringBuilder("new ").append(name).append("(");

         boolean gotone = false;

         // handle length
         if (primitive.hasArgument("length")) {
            String length = replaceDollars(primitive.getArgumentValue("length"));
            if (gotone)
               constructor.append(", ").append(length);
            else
               constructor.append(length);
            gotone = true;
         }

         // handle value
         if (primitive.hasArgument("value")) {
            String value = replaceDollars(primitive.getArgumentValue("value"));
            if (gotone)
               constructor.append(", ").append(value);
            else
               constructor.append(value);
            gotone = true;
         }

         constructor.append(")");

         // handle cast
         if (primitive.hasArgument("cast")) {
            String cast = primitive.getArgumentValue("cast");
            if (cast.equals("java.lang.String"))
               cast = "string";
            else if (cast.equals("java.lang.StringBuilder"))
               cast = "stringbuilder";
            else if (cast.equals("java.lang.StringBuffer"))
               cast = "stringbuffer";
            cast = cast.toLowerCase();
            constructor = new StringBuilder("new ").append(name).append("_cast_").append(cast).append("(").append(constructor).append(")");
         }

         return constructor.toString();
      } else if (item.isList()) {
         ListRuleItem list = item.getListSelf();
         String innerType = list.getInnerItem().getType().getType();
         String wrappedType = ListRuleItem.wrapType(innerType);
         return "new _yy_list<" + wrappedType + ">(" + replaceDollars(list.getLengthValue()) + ", " + parserConstructor(list.getInnerItem()) + ")";
      } else if (item.isSubparse()) {
         SubparseRuleItem subparse = item.getSubparseSelf();
         return "(new " + this.className + "(" + replaceDollars(subparse.getInputString()) + "))." + parserConstructor(subparse.getRule());
      } else {
         throw new RuntimeException("Unrecognized ruleitem: " + item);
      }
   }

   private void printStart(Rule start, PrintStream out) {
      Type type = start.getType();
      String returnType;
      if (type.isSimple()) {
         returnType = type.getSimpleSelf().getType();
      } else {
         returnType = type.getFunctionSelf().getReturnType().getType();
      }
      String wrappedType = ListRuleItem.wrapType(returnType);

      out.println(TAB + "public " + returnType + " parse() throws " + IOEXCEPTION + ", " + PARSEEXCEPTION + " {");
      out.println(TAB2 + RESULT + "<" + wrappedType + "> result = _yy_rule_" + start.getName() + ".get();");
      out.println(TAB2 + "if (result.failed())");
      out.println(TAB3 + "throw new " + PARSEEXCEPTION + "(\"Error parsing stream: \" + _yy_getResultMessage(result));");
      out.println(TAB2 + "else");
      out.println(TAB3 + "return result.result();");
      out.println(TAB + "}");
   }

   private String replaceDollars(String code) {
      int index = 0;
      while (index < code.length()) {
         if (code.charAt(index) == '$') {
            int start = ++index;
            char c;
            int number = 0;
            while (index<code.length() && '0' <= (c=code.charAt(index)) && c <= '9') {
               index++;
               number = number*10 + (c-'0');
            }
            if (index==start)
               continue;
            
            code = code.substring(0, start-1) + "_yy_" + number + code.substring(index);
            index = start;
         } else
            index++;
      }
      return code;
   }
}
