package pigbarf;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

public class MetaParser {
	private final DocumentBuilder builder;

	public MetaParser(DocumentBuilder _builder) {
		this.builder = _builder;
	}
	public MetaParser() throws ParserConfigurationException {
		this.builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
	}
	
	public Specification process(InputStream in) throws MetaParsingException {
		Document doc;
		try {
			doc = builder.parse(in);
		} catch (Throwable t) {
			throw new MetaParsingException("Cannot parse XML", t);
		}
		return this.processSpec(doc.getDocumentElement());
	}
	private Specification processSpec(Element spec) throws MetaParsingException {
		verify(spec, "spec", null, null);
		
		Specification specification = new Specification();
		Map<String,Type> typeMap = new HashMap<String,Type>();
		
		boolean hasTypes = false;
		boolean hasByteOrder = false;
		boolean hasBitOrder = false;
		boolean hasRules = false;
		boolean hasHeaders = false;
		
		List<Element> children = getElementChildren(spec);
		for (Element child : children) {
			final String childtag = child.getTagName();
			if (childtag.equals("types")) {
				if (hasTypes)
					throw new MetaParsingException("Multiple 'types' elements");
//				if (hasRules)
//					throw new MetaParsingException("Types element must come before rules element");
				processTypes(child, typeMap);
				hasTypes = true;
			}
			else if (childtag.equals("rules")) {
				if (hasRules)
					throw new MetaParsingException("Multiple 'rules' elements");
				processRules(child, specification);
				hasRules = true;
			}
			else if (childtag.equals("byte-order")) {
				if (hasByteOrder)
					throw new MetaParsingException("Multiple 'byte-order' elements");
				specification.setByteOrder(processByteOrder(child));
				hasByteOrder = true;
			}
			else if (childtag.equals("bit-order")) {
				if (hasBitOrder)
					throw new MetaParsingException("Multiple 'bit-order' elements");
				specification.setBitOrder(processBitOrder(child));
				hasBitOrder = true;
			}
			else if (childtag.equals("headers")) {
				if (hasHeaders)
					throw new MetaParsingException("Multiple 'headers' elements");
				specification.setHeaders(processHeaders(child));
			}
			else {
				throw new MetaParsingException("Unrecognied element: " + childtag);
			}
		}
		
		if (!hasRules)
			throw new MetaParsingException("Missing rules element");
		if (!hasTypes)
			throw new MetaParsingException("Missing types element");
		
		// this version will not do any type inference, for simplicity
		for (int i = 0; i < specification.getNumRules(); i++) {
			Rule rule = specification.getRule(i);
         
			if (!typeMap.containsKey(rule.getName()))
				throw new MetaParsingException("Rule " + rule.getName() + " has no type");
			Type type = typeMap.get(rule.getName());

			if (rule.getNumParameters() == 0) {
				if (!type.isSimple())
					throw new MetaParsingException("Simple rule has non-simple type: " + rule.getName());
				rule.setType(type);
			} else {
				if (!type.isFunction())
					throw new MetaParsingException("Parameterized rule has simple type: " + rule.getName());
				FunctionType func = type.getFunctionSelf();
				if (func.getNumParameters() != rule.getNumParameters())
					throw new MetaParsingException("Number of formals does not match type: " + rule.getName());
				rule.setType(type);
			}
		}

		// now resolve nonterminals (also inside of lists)
		for (int i = 0; i < specification.getNumRules(); i++) {
			Rule rule = specification.getRule(i);
         for (int k = 0; k < rule.getNumOptions(); k++) {
            RHS rhs = rule.getOption(k);
            for (int j = 0; j < rhs.getNumRuleItems(); j++) {
               RuleItem item = rhs.getRuleItem(j);
               if (!item.isValue()) continue;
               resolveValueItem(item.getValueSelf(), specification);
            }
         }
		}
		
		// now check for typed rules that do not exist
		for (String rulename : typeMap.keySet()) {
			if (specification.getRuleByName(rulename) == null)
				throw new MetaParsingException("Type give for nonexistent rule: " + rulename);
		}
		
		return specification;
	}
	
	private void resolveValueItem(ValueRuleItem valueitem, Specification spec)
	throws MetaParsingException {
		if (valueitem.isNonterminal()) {
			NonterminalRuleItem nonterm = valueitem.getNonterminalSelf();
			Rule callrule = spec.getRuleByName(nonterm.getName());
			if (callrule == null)
				throw new MetaParsingException("Nonterminal names nonexistent rule: " + nonterm.getName());
			nonterm.setRule(callrule);
		} else if (valueitem.isList()) {
			ListRuleItem list = valueitem.getListSelf();
			resolveValueItem(list.getInnerItem(), spec);
		} else if (valueitem.isSwitch()) {
         SwitchRuleItem s = valueitem.getSwitchSelf();
         ValueRuleItem defaultCase = s.getDefaultCase();
         SimpleType type = null;
         if (defaultCase != null) {
            resolveValueItem(defaultCase, spec);
            type = defaultCase.getType();
         }
         for (int i = 0; i < s.getNumCases(); i++) {
            SwitchRuleItem.Case c = s.getCase(i);
            resolveValueItem(c.rule, spec);
            if (type != null && !c.rule.getType().equals(type))
               throw new MetaParsingException("Types do not agree for switch rule item: " + type + " != " + c.rule.getType());
            type = c.rule.getType();
         }
      } else if (valueitem.isSubparse()) {
         SubparseRuleItem subparse = valueitem.getSubparseSelf();
         resolveValueItem(subparse.getRule(), spec);
      }
	}
			
	private String processHeaders(Element headers) throws MetaParsingException {
		verify(headers, "headers", null, null);
		return getChildText(headers).toString();
	}
	
	private void processTypes(Element types, Map<String,Type> typeMap) 
	throws MetaParsingException {
		verify(types, "types", null, null);
		StringBuilder text = getChildText(types);
		Buffer buffer = new Buffer(text);
		TypeParser parser = new TypeParser(buffer);
		parser.parseTypedecls(typeMap);
	}
	
	private void processRules(Element rules, Specification spec)  
	throws MetaParsingException {
		verify(rules, "rules", new String[]{"start"}, null);
		String start = rules.getAttribute("start");
		Buffer buffer = new Buffer(getChildText(rules));
		RuleParser parser = new RuleParser(buffer);
		parser.parseRules(spec);
		
		Rule startRule = spec.getRuleByName(start);
		if (startRule == null)
			throw new MetaParsingException("Start rule " + start + " does not exist");
		if (startRule.getNumParameters() != 0)
			throw new MetaParsingException("Start rule cannot be parameterized");
		spec.setStartRule(startRule);
	}
	
	private boolean processByteOrder(Element byteOrder) 
	throws MetaParsingException {
		verify(byteOrder, "byte-order", new String[]{"value"}, null);
		String order = byteOrder.getAttribute("value");
		if (order.equals("little"))
			return false;
		else if (order.equals("big"))
			return true;
		else
			throw new MetaParsingException("Element byte-order must have 'little' or 'big' as its value attribute");
	}
	
	private boolean processBitOrder(Element bitOrder) 
	throws MetaParsingException {
		verify(bitOrder, "bit-order", new String[]{"value"}, null);
		String order = bitOrder.getAttribute("value");
		if (order.equals("little"))
			return false;
		else if (order.equals("big"))
			return true;
		else
			throw new MetaParsingException("Element bit-order must have 'little' or 'big' as its value attribute");
	}
	
	/////////////////////////////////////////
	
	private StringBuilder getChildText(Element elt) {
		StringBuilder builder = new StringBuilder();
		NodeList children = elt.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (child instanceof Text) {
				builder.append(((Text)child).getData());
			}
			else if (child instanceof CDATASection) {
				builder.append(((CDATASection)child).getData());
			}
		}
		return builder;
	}
	
	private List<Element> getElementChildren(Element elt) {
		List<Element> result = new ArrayList<Element>();
		NodeList children = elt.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (child instanceof Element)
				result.add((Element)child);
		}
		return result;
	}
	
	private void verify(
			Element elt, 
			String tag, 
			String[] required, 
			String[] optional) throws MetaParsingException {
		if (!elt.getTagName().equals(tag))
			throw new MetaParsingException("Element has wrong tag name. Expected \"" + tag + "\", found \"" + elt.getTagName() + "\"");
		NamedNodeMap attrs = elt.getAttributes();
		Set<String> allowed = new HashSet<String>();
		if (required != null)
			allowed.addAll(Arrays.asList(required));
		if (optional != null)
			allowed.addAll(Arrays.asList(optional));
		for (int i = 0; i < attrs.getLength(); i++) {
			Attr attr = (Attr)attrs.item(i);
			if (!allowed.contains(attr.getName()))
				throw new MetaParsingException("Attribute not allowed for " + tag + " element: " + attr.getName());
		}
		if (required != null) {
			for (String req : required) {
				if (!elt.hasAttribute(req))
					throw new MetaParsingException("Element " + tag + " is missing required attribute: " + req);
			}
		}
	}
}