package pigbarf;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public abstract class PrimitiveAmbassador {
	public static void checkArgs(Primitive prim, Map<String,String> args) 
	throws MetaParsingException {
		switch (prim) {
		case BIT: checkBit(args); break;
		case BITS: checkBits(args); break;
		case BYTE: checkByte(args); break;
		case BYTES: checkBytes(args); break;
		case CHAR: checkChar(args); break;
		case CHARS: checkChars(args); break;
		case DOUBLE: checkDouble(args); break;
		case FLOAT: checkFloat(args); break;
		case INT: checkInt(args); break;
		case LONG: checkLong(args); break;
		case SHORT: checkShort(args); break;
		default:
			throw new MetaParsingException("Unknown primitive: " + prim.name());
		}
	}

	private static void checkBit(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 * 	cast -> byte, int, short, long
       *    value -> actual
		 */
		verify("bit", args, null, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("byte") || cast.equals("int") || cast.equals("short") || cast.equals("long")))
				throw new MetaParsingException("Primitive bit's cast argument must be one of byte|short|int|long");
		}
	}
	private static void checkBits(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 * 	cast -> byte, int, short, long
		 * 	length -> actual
		 *    value -> actual
		 */
		verify("bits", args, new String[]{"length"}, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("byte") || cast.equals("int") || cast.equals("short") || cast.equals("long")))
				throw new MetaParsingException("Primitive bits may only have cast value of byte|short|int|long");
		}
	}
	private static void checkByte(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 * 	cast -> int, short, long
		 *    value -> actual
		 */
		verify("byte", args, null, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("int") || cast.equals("short") || cast.equals("long")))
				throw new MetaParsingException("Primitive byte's cast argument must be one of short|int|long");
		}
	}
	private static void checkBytes(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 * 	cast -> int, short, long
		 * 	length -> actual
       *    value -> actual
		 */
		verify("bytes", args, new String[]{"length"}, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("int") || cast.equals("short") || cast.equals("long")))
				throw new MetaParsingException("Primitive bytes may only have cast value of short|int|long");
		}
	}
	private static void checkChar(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 *    value -> actual
		 */
		verify("char", args, null, new String[]{"value"});
	}
	private static void checkChars(Map<String,String> args) throws MetaParsingException {
		/* allows: 
		 * 	cast -> String | StringBuilder | StringBuffer
		 * 	length -> actual
		 *    value -> actual
		 */
		verify("chars", args, new String[]{"length"}, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("String") || 
               cast.equals("java.lang.String") ||
               cast.equals("StringBuilder") ||
               cast.equals("java.lang.StringBuilder") ||
               cast.equals("StringBuffer") ||
               cast.equals("java.lang.StringBuffer")))
				throw new MetaParsingException("Primitive chars may only have cast value of String|StringBuilder|StringBuffer");
		}
	}
	private static void checkDouble(Map<String,String> args) throws MetaParsingException {
		/* allows:
		 *    value -> actual
		 */
		verify("double", args, null, new String[]{"value"});
	}
	private static void checkFloat(Map<String,String> args) throws MetaParsingException {
		/* allows:
		 *    cast -> double
		 *    value -> actual
		 */
		verify("float", args, null, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!cast.equals("double"))
				throw new MetaParsingException("Primitive float may only have cast value of double");
		}
	}
	private static void checkInt(Map<String,String> args) throws MetaParsingException {
		/* allows:
		 *    cast -> long
		 *    value -> actual
		 */
		verify("int", args, null, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!cast.equals("long"))
				throw new MetaParsingException("Primitive int may only have cast value of long");
		}
	}
	private static void checkLong(Map<String,String> args) throws MetaParsingException {
		/* allows:
		 *    value -> actual
		 */
		verify("long", args, null, new String[]{"value"});
	}
	private static void checkShort(Map<String,String> args) throws MetaParsingException {
		/* allows:
		 *    cast -> int, long
		 *    value -> actual
		 */
		verify("short", args, null, new String[]{"cast", "value"});
		if (args.containsKey("cast")) {
			String cast = args.get("cast");
			if (!(cast.equals("long") || cast.equals("int")))
				throw new MetaParsingException("Primitive short may only have cast value of int|long");
		}
	}

	private static void verify(
			String name,
			Map<String,String> args, 
			String[] required, 
			String[] optional) throws MetaParsingException {
		Set<String> allowed = new HashSet<String>();
		if (required != null)
			allowed.addAll(Arrays.asList(required));
		if (optional != null)
			allowed.addAll(Arrays.asList(optional));
		for (String key : args.keySet()) {
			if (!allowed.contains(key))
				throw new MetaParsingException("Primitive " + name + " does not allow argument " + key);
		}
		if (required != null) {
			for (String req : required) {
				if (!args.containsKey(req))
					throw new MetaParsingException("Primitive " + name + " is missing required argument " + req);
			}
		}
	}
}
