package pigbarf;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/*
 * rule ::= ident [ formals ] '::=' rhs ';;'
 *
 * rhs ::= ruleitem+ | ruleitem+ '|' rhs
 * 
 * formals ::= '(' ident (',' ident)* ')'
 * 
 * ruleitem ::= value | predicate | code | align | eof
 *
 * eof ::= 'EOF'
 *
 * align ::= 'align' '(' code ')'
 * 
 * value ::= nonterminal | primitive | list | switch | subparse | expression
 *
 * expression ::= '[:[' typestr ']:[' javacode ']:]'
 *
 * typestr ::= uninterpreted java string, specifies a type (anything without ']:[')
 *
 * subparse ::= 'subparse' '(' ( inputpair ';' typepair | typepair ';' inputpair ) ')'
 *
 * inputpair ::= 'input' '=' code
 *
 * switch ::= 'switch' code '{' case+ [ default_case ] '}'
 *
 * case ::= code '->' value ';;'
 *
 * default_case ::= 'default' '->' value ';;'
 *
 * list ::= 'list' '(' ( lengthpair ';' typepair  | typepair ';' lengthpair ) ')'
 * 
 * nonterminal ::= ident [ actuals ]
 *
 * actuals ::= '(' code (';' code)* ')'
 * 
 * primitive ::= 'bit'    [ namedargs ]
 * 	 		   | 'bits'   [ namedargs ] 
 * 			   | 'byte'   [ namedargs ] 
 * 			   | 'bytes'  [ namedargs ]
 * 	 		   | 'int'    [ namedargs ]
 * 	         | 'short'  [ namedargs ] 
 *             | 'long'   [ namedargs ]
 * 			   | 'char'   [ namedargs ] 
 * 			   | 'chars'  [ namedargs ]
 * 			   | 'float'  [ namedargs ]
 * 			   | 'double' [ namedargs ] 
 * 
 * namedargs ::= '(' keyvaluepair (';' keyvaluepair)* ')'
 * 
 * keyvaluepair ::= lengthpair | 'value' '=' code | 'cast' '=' ident
 * 
 * lengthpair ::= 'length' '=' code
 *
 * typepair ::= 'type' '=' value
 * 
 * predicate ::= '{{' javapred '}}'
 * 
 * javapred ::= uninterpreted java boolean expression, anything without '}}'
 * 
 * code ::= '[[' javacode ']]'
 * 
 * javacode ::= uninterpreted java code, anything without ']]'
 */
public class RuleParser {
   private static final boolean DEBUG = false;
   private static void debug(String msg) {
      if (DEBUG)
         System.err.println("RuleParser: " + msg);
   }

	private final Buffer buffer;
	
	public RuleParser(Buffer _buffer) {
		this.buffer = _buffer;
	}
	
	public void parseRules(Specification spec) throws MetaParsingException {
		buffer.skipWS();
		while (!buffer.atEOF()) {
			Rule rule = parseRule();
			if (spec.getRuleByName(rule.getName()) != null)
				throw new MetaParsingException("Duplicate rules specified: " + rule.getName());
			spec.addRule(rule);
			buffer.skipWS();
		}
	}
	
	public Rule parseRule() throws MetaParsingException {
		// rule ::= ident [ formals ] '::=' rhs ';;'
		String ruleName = parseIdent();

      debug("rule name: " + ruleName);

		List<String> formals = new ArrayList<String>();
		buffer.skipWS();
		if (buffer.next("(")) {
			// formals ::= '(' ident (',' ident)* ')'
			buffer.get();
			formals.add(parseIdent());
			buffer.skipWS();
			while (buffer.next(",")) {
				buffer.get();
				String formal = parseIdent();
				if (formals.contains(formal))
					throw new MetaParsingException("Duplicate formal parameter name: " + formal);
				formals.add(formal);
				buffer.skipWS();
			}
			buffer.skipWS();
         must(")");
		}
		
		buffer.skipWS();
      must("::=");
		
		Rule rule = new Rule(ruleName, formals);
		
      // now parse options
      RHS currentRHS = new RHS();

		buffer.skipWS();
		while (!buffer.next(";;")) {
         if (buffer.next("|")) {
            buffer.get();
            if (currentRHS.getNumRuleItems() == 0)
               throw new MetaParsingException("Option has no rule items");
            rule.addOption(currentRHS);
            currentRHS = new RHS();
         } else {
            currentRHS.addRuleItem(parseRuleItem());
         }
			buffer.skipWS();
		}
      must(";;");
      if (currentRHS.getNumRuleItems() == 0)
         throw new MetaParsingException("Option has no rule items");
      rule.addOption(currentRHS);
		
		return rule;
	}
	
	public ValueRuleItem parseValue() throws MetaParsingException {
		// value ::= nonterminal | primitive | list | switch | subparse | expression

      buffer.skipWS();
      if (buffer.next("[:[")) {
         // expression ::= '[:[' typestr ']:[' javacode ']:]'
         must("[:[");
         StringBuilder typestr = new StringBuilder();
         while (!buffer.atEOF() && !buffer.next("]:[")) {
            typestr.append(buffer.get());
         }
         mustWS("]:[");
         StringBuilder expression = new StringBuilder();
         while (!buffer.atEOF() && !buffer.next("]:]")) {
            expression.append(buffer.get());
         }
         mustWS("]:]");
         return new ExpressionRuleItem(new SimpleType(typestr.toString()), expression.toString());
      } else {
         String ident = parseIdent(); // will skip ws
         Primitive prim = Primitive.getByName(ident);
         if (prim != null) {
            /* primitive ::= 'bit' [ namedargs ]
             * 	 		   | 'bits' [ namedargs ] 
             * 			   | 'byte' [ namedargs ] 
             * 			   | 'bytes' [ namedargs ]
             * 	 		   | 'int' [ namedargs ]
             * 		      | 'short' [ namedargs ] 
             * 		      | 'long' [ namedargs ]
             * 			   | 'char' [ namedargs ] 
             * 			   | 'chars' [ namedargs ]
             * 			   | 'float' [ namedargs ]
             * 			   | 'double' [ namedargs ] 
             */
            buffer.skipWS();
            Map<String,String> namedargs = new HashMap<String,String>();
            if (buffer.next("("))
               parseNamedArgs(namedargs);
            return new PrimitiveRuleItem(prim, namedargs);
         } else if (ident.equals("list")) {
            // list ::= 'list' '(' 'type' '=' value ';' 'length' '=' actual ')'
            buffer.skipWS();
            must("(");

            Map<String,String> args = new HashMap<String,String>();
            buffer.skipWS();
            ValueRuleItem inner;
            if (buffer.next("type")) {
               inner = parseTypePair();
               buffer.skipWS();
               must(";");
               parseLengthPair(args);
            } else if (buffer.next("length")) {
               parseLengthPair(args);
               buffer.skipWS();
               must(";");
               inner = parseTypePair();
            } else {
               throw new MetaParsingException("Expecting 'type' or 'length'");
            }

            buffer.skipWS();
            must(")");
			
            return new ListRuleItem(inner, args.get("length"));
         } else if (ident.equals("switch")) {
            // switch ::= 'switch' code '{' case+ [ default_case ] '}'
            buffer.skipWS();
            String target = parseCode();
            buffer.skipWS();
            must("{");
         
            ValueRuleItem defaultCase = null;
            List<SwitchRuleItem.Case> cases = new ArrayList<SwitchRuleItem.Case>();

            buffer.skipWS();
            while (!buffer.next("}")) {
               if (buffer.next("default")) {
                  // default_case ::= 'default' '->' value ';;'
                  if (defaultCase != null)
                     throw new MetaParsingException("Multiple default cases specified for switch");
                  must("default");
                  buffer.skipWS();
                  must("->");
                  defaultCase = parseValue();
                  buffer.skipWS();
                  must(";;");
               } else {
                  // case ::= code '->' value ';;'
                  String match = parseCode();
                  buffer.skipWS();
                  must("->");
                  ValueRuleItem rule = parseValue();
                  buffer.skipWS();
                  must(";;");
                  cases.add(new SwitchRuleItem.Case(match, rule));
               }

               buffer.skipWS();
            }
            must("}");
         
            if (cases.size() == 0)
               throw new MetaParsingException("Switch rule must have at least 1 case");

            return new SwitchRuleItem(target, cases, defaultCase);
         } else if (ident.equals("subparse")) {
            // subparse ::= 'subparse' '(' ( inputpair ';' typepair | typepair ';' inputpair ) ')'
            //
            // inputpair ::= 'input' '=' code
            mustWS("(");
            ValueRuleItem type;
            String input;

            String first = parseIdent();
            if (first.equals("type")) {
               mustWS("=");
               type = parseValue();
               mustWS(";");
               if (!parseIdent().equals("input"))
                  throw new MetaParsingException("Expecting 'input'");
               mustWS("=");
               input = parseCode();
            } else if (first.equals("input")) {
               mustWS("=");
               input = parseCode();
               mustWS(";");
               if (!parseIdent().equals("type"))
                  throw new MetaParsingException("Expecting 'type'");
               mustWS("=");
               type = parseValue();
            } else {
               throw new MetaParsingException("Expecting 'input'/'type'");
            }
            mustWS(")");

            if (type instanceof SwitchRuleItem || type instanceof SubparseRuleItem)
               throw new MetaParsingException("Type value must not be subparse or switch");

            return new SubparseRuleItem(input, type);
         } else {
            /* nonterminal ::= ident [ actuals ]
             * 
             * actuals ::= '(' actual (';' actual)* ')'
             */
            buffer.skipWS();
            List<String> actuals = new ArrayList<String>();
            if (buffer.next("(")) {
               buffer.get();
               actuals.add(parseCode());
               buffer.skipWS();
               while (buffer.next(";")) {
                  buffer.get();
                  actuals.add(parseCode());
                  buffer.skipWS();
               }
               must(")");
            }
            return new NonterminalRuleItem(ident, actuals);
         }
      } 
	}
	
	public RuleItem parseRuleItem() throws MetaParsingException {
		// ruleitem ::= nonterminal | primitive | predicate | code | switch | eof
		
      debug("ruleitem peek: " + buffer.peekN(5));

		buffer.skipWS();
		if (buffer.next("{{")) {
			/* predicate ::= '{{' javapred '}}'
			 * 
			 * javapred ::= uninterpreted java code, anything without '}}'
			 */
			buffer.getN(2);
			StringBuilder builder = new StringBuilder();
			while (buffer.hasN(1) && !buffer.next("}}")) {
				builder.append(buffer.get());
			}
         must("}}");

         PredicateRuleItem result = new PredicateRuleItem(builder.toString());
         debug("parsed predicate: " + result);
			return result;
		} 
		else if (buffer.next("[[")) {
			/* code ::= '[[' javacode ']]'
			 * 
			 * javacode ::= uninterpreted java code, anything without ']]'
			 */
         CodeRuleItem result = new CodeRuleItem(parseCode());
         debug("parsed code: " + result);
         return result;
		} 
      else {
         int pos = buffer.getIndex();
         String ident = parseIdent();
         buffer.setIndex(pos);
         if (ident.equals("align")) {
            AlignRuleItem result = parseAlign();
            debug("parsed align: " + result);
            return result;
         } else if (ident.equals("EOF")) {
            must("EOF");
            return EOFRuleItem.INSTANCE;
         } else {
            ValueRuleItem result = parseValue();
            debug("parsed value: " + result);
            return result;
         }
      }
	}

   public AlignRuleItem parseAlign() throws MetaParsingException {
      // align ::= 'align' '(' code ')'      
      buffer.skipWS();
      must("align");
      buffer.skipWS();
      must("(");
      String code = parseCode();
      buffer.skipWS();
      must(")");
      return new AlignRuleItem(code);
   }

   public String parseCode() throws MetaParsingException {
      /* code ::= '[[' javacode ']]'
       * 
       * javacode ::= uninterpreted java code, anything without ']]'
       */
      buffer.skipWS();
      must("[[");
      StringBuilder builder = new StringBuilder();
      while (buffer.hasN(1) && !buffer.next("]]")) {
         builder.append(buffer.get());
      }
      must("]]");
      return builder.toString();
   }

	public ValueRuleItem parseTypePair() 
	throws MetaParsingException {
      // typepair ::= 'type' '=' value
      mustWS("type");
      mustWS("=");
      return parseValue();
   }

	public void parseLengthPair(Map<String,String> args) 
	throws MetaParsingException {
      // lengthpair ::= 'length' '=' code
      buffer.skipWS();
      must("length");
      buffer.skipWS();
      must("=");
      String code = parseCode();
      if (args.containsKey("length"))
         throw new MetaParsingException("Duplicate arg: length");
      args.put("length", code);
   }

   public void parseKeyvaluePair(Map<String,String> args) 
	throws MetaParsingException {
      /* keyvaluepair ::= lengthpair | 'value' '=' code | 'cast' '=' ident
       * 
       * lengthpair ::= 'length' '=' code
       */
      buffer.skipWS();
      if (buffer.next("length")) {
         parseLengthPair(args);
      } 
      else if (buffer.next("cast")) {
         buffer.getN(4);
         buffer.skipWS();
         must("=");
         String cast = parseIdent();
         if (args.containsKey("cast"))
            throw new MetaParsingException("Duplicate arg: cast");
         args.put("cast", cast);
      }
      else if (buffer.next("value")) {
         buffer.getN(5);
         buffer.skipWS();
         must("=");
         String code = parseCode();
         if (args.containsKey("value"))
            throw new MetaParsingException("Duplicate arg: value");
         args.put("value", code);
      } 
      else 
         throw new MetaParsingException("Expecting 'length' or 'cast' or 'value'");
   }
	
	public void parseNamedArgs(Map<String,String> args) 
	throws MetaParsingException {
      /* namedargs ::= '(' keyvaluepair (';' keyvaluepair)* ')'
       */
		buffer.skipWS();
      must("(");
      parseKeyvaluePair(args);
      buffer.skipWS();

      while (buffer.next(";")) {
         buffer.get();
         parseKeyvaluePair(args);
         buffer.skipWS();
      }
      must(")");
	}

	public String parseIdent() throws MetaParsingException {
		StringBuilder builder = new StringBuilder();
		buffer.skipWS();
		loop:
		while (!buffer.atEOF()) {
			char c = buffer.peek();
			if ('a' <= c && c <= 'z')
				builder.append(c);
			else if ('A' <= c && c <= 'Z')
				builder.append(c);
			else if ('0' <= c && c <= '9')
				builder.append(c);
			else {
				switch (c) {
				case '_':
					builder.append(c);
					break;
				
				default:
					break loop;
				}
			}
			buffer.get();
		}
		if (builder.length() == 0)
			throw new MetaParsingException("Identifier had 0 length");
		return builder.toString();
	}

   private void must(String str) throws MetaParsingException {
      if (!buffer.next(str))
         throw new MetaParsingException("Expecting '" + str + "'");
      buffer.getN(str.length());
   }
   private void mustWS(String str) throws MetaParsingException {
      buffer.skipWS();
      if (!buffer.next(str))
         throw new MetaParsingException("Expecting '" + str + "'");
      buffer.getN(str.length());
   }
}
