package pigbarf;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/*
 * typedecl ::= type ':' rulenamelist ';'
 * 
 * rulenamelist ::= rulename | rulename ',' rulenamelist
 *
 * rulename ::= valid rule name identifier
 * 
 * type ::= function | simple
 * 
 * function ::= (simplelist) simple
 * 
 * simplelist ::= simple | simple ',' simplelist
 * 
 * simple ::= anything without ',' or '(' or ')'
 */
public class TypeParser {
	private final Buffer buffer;
	
	public TypeParser(Buffer _buffer) {
		this.buffer = _buffer;
	}
	
	public void parseTypedecls(Map<String,Type> typeMap)
	throws MetaParsingException {
		buffer.skipWS();
		while (!buffer.atEOF()) {
			Type type = parseType();

			buffer.skipWS();
			if (!(buffer.hasN(1) && buffer.get() == ':'))
				throw new MetaParsingException("Error parsing type decls: Expecting ':'");

			List<String> ruleNames = new ArrayList<String>();
			ruleNames.add(parseRulename());
			buffer.skipWS();
			while (buffer.hasN(1) && buffer.peek() == ',') {
				buffer.get();
				ruleNames.add(parseRulename());
				buffer.skipWS();
			}
			buffer.skipWS();
			if (!(buffer.hasN(1) && buffer.get()==';'))
				throw new MetaParsingException("Error parsing type decls: Expecting ';'");
         buffer.skipWS();
			
			for (String rulename : ruleNames) {
				if (typeMap.containsKey(rulename))
					throw new MetaParsingException("Error parsing type decls: Rule " + rulename + " has multiple type declarations");
				typeMap.put(rulename, type);
			}
		}
	}
	
	private String parseRulename() throws MetaParsingException {
		StringBuilder builder = new StringBuilder();
		buffer.skipWS();
		loop:
		while (!buffer.atEOF()) {
			char c = buffer.peek();
			if ('a' <= c && c <= 'z') {
				builder.append(c);
			} else if ('A' <= c && c <= 'Z') {
				builder.append(c);
			} else if ('0' <= c && c <= '9') {
				builder.append(c);
			} else {
				switch (c) {
				case '_':
				case '#':
				case '!':
				case '@':
				case '%':
				case '^':
				case '&':
				case '*':
				case '+':
				case '=':
				case '-':
					builder.append(c);
					break;
				
				default:
					break loop;
				}
			}
			buffer.get();
		}
		if (builder.length() == 0)
			throw new MetaParsingException("Error parsing type decls: No rulename chars to read");
		return builder.toString();
	}
	
	private SimpleType parseSimple() throws MetaParsingException {
		buffer.skipWS();
		StringBuilder builder = new StringBuilder();
		loop:
		while (!buffer.atEOF()) {
			char c = buffer.peek();
			switch (c) {
			case '(':
			case ')':
			case ',':
			case ':':
			case ';':
				break loop;
			default:
				builder.append(c);
            buffer.get();
				break;
			}
		}
      String type = builder.toString().trim();
		if (type.length() == 0)
			throw new MetaParsingException("Error parsing type decls: No simple type chars to read");
		return new SimpleType(type);
	}

	private Type parseType() throws MetaParsingException {
		buffer.skipWS();
		char first = buffer.peek();
		if (first == '(') {
			/*
			 * function ::= (simplelist) simple
			 * 
			 * simplelist ::= simple | simple ',' simplelist
			 */
			buffer.get();
			List<SimpleType> params = new ArrayList<SimpleType>();
			params.add(parseSimple());
			buffer.skipWS();
			while (buffer.hasN(1) && buffer.peek()==',') {
				buffer.get();
				params.add(parseSimple());
				buffer.skipWS();
			}
			buffer.skipWS();
			if (!(buffer.hasN(1) && buffer.get() == ')'))
				throw new MetaParsingException("Error parsing type decls: Expecting ')'");
			SimpleType returnType = parseSimple();
			return new FunctionType(returnType, params);
		} 
		else {
			// simple
			return parseSimple();
		}
	}
}
